# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
# {xrst_begin newline_indices dev}
# {xrst_comment_ch #}
#
# Find index of all the newlines in a string
# ##########################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data
# ****
# The string we are searching for newlines.
#
# Results
# *******
#
# newline_list
# ************
# The return newline_list is the list of indices in data that
# represent all of the newlines; i.e. '\n'.
#
# {xrst_end newline_indices}
# BEGIN_DEF
def newline_indices(data) :
   assert type(data) == str
   # END_DEF
   pattern_newline  = re.compile( r'\n')
   newline_itr      = pattern_newline.finditer(data)
   newline_list     = list()
   for m_obj in newline_itr :
      next_index = m_obj.start()
      newline_list.append( next_index )
   # BEGIN_RETURN
   #
   assert type(newline_list) == list
   if 0 < len( newline_list) :
      assert type(newline_list[0]) == int
   return newline_list
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import xrst
# {xrst_begin next_heading dev}
# {xrst_comment_ch #}
#
# Return location of the next heading in a page
# #############################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data
# ****
# is the data that we are searching for a heading in. The heading text must
# have at least one character and be followed by an underline of at least the
# same length. The heading text may be proceeded by an overline.
#
# data_index
# **********
# is the index in the data where the search starts. This must be zero
# or directly after a newline.
#
# page_file
# *********
# name of the file that contains the input begin command for this page.
# This is only used for error reporting.
#
# page_name
# *********
# is the name of this page.
# This is only used for error reporting.
#
# heading_index
# *************
# If there is an overline, this is the index in data of the beginning of the
# overline. Otherwise, it is the index of the beginning of the heading text.
# If 0 < heading_index, there is a newline just before heading_index; i.e.,
# data[heading_index]=='\n'.  If heading_index is -1, there is no heading
# in data that begins at or after data_index.
#
# heading_text
# ************
# if 0 <= heading_index, this is the heading text.
#
# underline_text
# **************
# if 0 <= heading_index, this is the underline text.
# If there is an overline present, it is the same as the underline text.
#
# {xrst_end next_heading}
# BEGIN_DEF
def next_heading(data, data_index, page_file, page_name) :
   assert type(data) == str
   assert type(data_index) == int
   if data_index != 0 :
      assert data[data_index-1] == '\n'
   assert type(page_file) == str
   assert type(page_name) == str
   # END_DEF
   #
   # punctuation
   punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
   assert len(punctuation) == 34 - 2 # two escape sequences
   #
   # next_start, next_newline
   next_start       = data_index
   next_newline     = data.find('\n', next_start)
   #
   # state
   state  = 'before_overline'
   #
   while 0 <= next_newline :
      #
      # next_line
      next_line = data[next_start : next_newline]
      next_line = xrst.pattern['line'].sub('', next_line)
      next_line = next_line.rstrip(' \t')
      #
      # next_len
      next_len  = len(next_line)
      #
      if next_len < 1 :
         # next_line cannot be an overline, heading, or underline
         state = 'before_overline'
      #
      elif state == 'before_overline' :
         #
         # heading_index
         heading_index = next_start
         #
         ch = next_line[0]
         if ch in punctuation and next_line == ch * next_len :
            # next_line can be an overline but not an underline
            state         = 'before_heading'
            overline_text = next_line
         else :
            # next_line can be a heading, but not overline or underline
            state         = 'before_underline'
            overline_text = None
            heading_text  = next_line
      #
      elif state == 'before_heading' :
         # next_line can be a heading
         state = 'before_underline'
         heading_text    = next_line
      #
      elif state == 'before_underline' :
         underline_ch = next_line[0]
         if underline_ch not in punctuation :
            state = 'before_overline'
         elif next_line != underline_ch * next_len :
            state = 'before_overline'
         elif len(next_line) < len(heading_text) :
            state = 'before_overline'
         else :
            m_line = xrst.pattern['line'].search(data, next_start)
            line   = m_line.group(1)
            msg = ''
            if len(heading_text) < len(next_line) :
               msg = 'underline text is longer than line above it'
            elif overline_text is not None and overline_text != next_line:
               msg = 'overline text is non-empty and not equal underline text'
            if msg != '' :
               xrst.system_exit(
                     msg,
                     file_name = page_file,
                     page_name = page_name,
                     m_obj     = m_line,
                     data      = data,
               )
            #
            # underline_text
            underline_text = next_line
            #
            return heading_index, heading_text, underline_text
      #
      # next_start, next_newline
      next_start   = next_newline + 1
      next_newline = data.find('\n', next_start)
   #
   # heading_index, heading_text, underline_text
   heading_index = -1
   heading_text  = ''
   underline_text  = ''
   #
   # BEGIN_RETURN
   #
   assert type(heading_index) == int
   assert type(heading_text) == str
   assert type(underline_text) == str
   return heading_index, heading_text, underline_text
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin template_cmd user}

Template Command
################

Syntax
******
| ``\{xrst_template`` *separator*
| |tab| *template_file*
| |tab| *match_1* *separator* *replace_1*
| |tab| *match_2* *separator* *replace_2*
| |tab| ...
| |tab| *comment_1*
| |tab| *comment_2*
| |tab| ...
| ``}``

Input File
**********
During the expansion of a template command,
the current input file is its *template_file* .
Otherwise, the current input file is the current :ref:`begin_cmd@Page File` .

Rst Include
***********
A similar sphinx include directive, in the context of xrst, is:

| |tab| .. include \{xrst_dir *file_name* }

see :ref:`dir_cmd-name` .
The template command differs form the include directive in the following ways:

#. The template command allows for text replacement
   during the include so that *template_file* is like function or macro.

#. It also allows for conditional including of sections of the template file
   when combined with the
   :ref:`suspend_cmd@boolean` or :ref:`suspend_cmd@left, right`
   arguments in the suspend command.

#. Errors and warnings in a template expansion will include both
   the line in the template file and the line where it is used.
   Errors and warnings in a sphinx include only report the
   line number in the include file.

#. xrst commands in *template_file* ( *file_name* )
   will get interpreted (will not get interpreted).

#. Template command :ref:`comments <template_cmd@comment>` can be used
   to check that the included file still satisfies the requirements
   of the file doing the include.

White Space
***********
The newline character separates the lines of input above
and excluded from white space in the discussion below..

separator
*********
The *separator* argument is a single character that separates
matches from their replacements.
Leading and trailing white space around *separator* is ignored.

template_file
*************
is the name of the template file.
Leading and trailing white space around *template_file* is ignored
and *template_file* cannot contain the ``@`` character
(the template file may contain the ``@`` character).
Template files are different from other xrst input file
because none of the following xrst commands can be in a template expansion:
:ref:`begin_cmd-name` ,
:ref:`comment_ch_cmd-name` ,
:ref:`indent_cmd-name` ,
:ref:`spell_cmd-name` ,
:ref:`template_cmd-name` .

match
*****
Each *match* in the template file gets replaced.
Leading and trailing white space around each *match* is ignored.

replace
*******
For each *match*, the corresponding *replace* is used in its place.
Leading and trailing white space around each *replace* is ignored.

comment
*******
A *comment* is any line below the *template_file* that
does not contain the *separator* character.
Leading and trailing white space around each *comment* is ignored.
If *comment* is empty, it is ignored.
Otherwise, it is an error if a *comment* does not appear in the template file
(before template expansion).
In other words, a template file can have a list of comments
that can be in a template command that uses the template file.
This enables the template command to check that certain features
of the template file have not changed.

Command End
***********
The first occurrence of a right brace ``}`` ,
directly after a newline ,
terminates the template command.

Example
*******
:ref:`template_example-name`

{xrst_end template_cmd}
"""
# ----------------------------------------------------------------------------
import os
import re
import xrst
#
#
# pattern_template:
# 1: character preceeding the template command
# 2. the separator (including surrounding white space)
# 3. line number where the template command appeared
# 4. rest of template command
pattern_template    = re.compile(
   r'([^\\]){xrst_template([^\n}]*)@xrst_line ([0-9]+)@\n([^}]*)}'
)
# pattern_arg
pattern_arg       = re.compile( r'([^\n]*)@xrst_line ([0-9]+)@\n| *\n' )
# ----------------------------------------------------------------------------
# {xrst_begin template_cmd_dev dev}
# {xrst_spell
# }
# {xrst_comment_ch #}
#
# Expand the template commands in a page
# ######################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# Restrictions
# ************
# The template expansion must come before processing any other commands
# except for the following:
# begin, end, comment_ch, indent, suspend, resume, spell, template.
#
# data_in
# *******
# is the data for a page before the
# :ref:`template commands <template_cmd-name>` have been expanded.
#
# page_file
# *********
# is the name of the file, for this page, where the begin command appears.
# This is used for error reporting .
#
# page_name
# *********
# is the name of the page that this data is in. This is only used
# for error reporting.
#
# data_out
# ********
# Each xrst template command is expanded and
# xrst.add_line_numbers is used to add line numbers corresponding to the
# template file.
# In addition, the following text is added at the beginning and end of the
# expansion:
#
# | |tab| @ ``\{xrst_template_begin`` @ *template_file* @ *page_line* @ ``}`` @
# | |tab| @ ``\{xrst_template_end}`` @
#
# where *page_line* is the line where the line number in *page_file*
# where the template command appeared. There is no white space between
# the tokens above.
#
#
# {xrst_end template_cmd_dev}
# BEGIN_DEF
def template_command(data_in, page_file, page_name) :
   assert type(data_in) == str
   assert type(page_file) == str
   assert type(page_name) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_template
   m_template  = pattern_template.search(data_out , 0)
   while m_template != None :
      #
      # separator
      separator = m_template.group(2).strip()
      if len(separator) != 1 :
         msg  =  '{xrst_template separator\n'
         msg += f'separator = "{separator}" must be one character'
         xrst.system_exit(
            msg,
            file_name = page_file,
            page_name = page_name,
            m_obj     = m_template,
            data      = data_out
         )
      #
      # arg_text
      # rest of template command
      arg_text = m_template.group(4)
      #
      # template_file, template_file_line, template_file_end
      template_file = ''
      m_arg         = pattern_arg.search(arg_text)
      if m_arg != None :
         if m_arg.group(1) != None :
            template_file =  m_arg.group(1).strip()
      if template_file == '' :
         msg = 'template command: the template_file is missing.'
         xrst.system_exit(
            msg,
            file_name = page_file,
            page_name = page_name,
            m_obj     = m_arg,
            data      = arg_text
         )
      if 0 < template_file.find('@')  :
         msg = 'template command: template_file contains the @ character.'
         xrst.system_exit(
            msg,
            file_name = page_file,
            page_name = page_name,
            m_obj     = m_arg,
            data      = arg_text
         )
      template_file_line = m_arg.group(2)
      template_file_end  = m_arg.end()
      #
      # template_data
      if not os.path.isfile(template_file) :
         msg  = f'template command: template_file = {template_file}\n'
         msg += 'can not find the template file'
         #
         # template command cannot be inside a template file so we can use
         # line number to report the error.
         xrst.system_exit(msg,
            file_name = page_file,
            page_name = page_name,
            line      = template_file_line,
         )
      file_obj       = open(template_file, 'r')
      template_data  = file_obj.read()
      file_obj.close()
      #
      # Check comments
      m_arg = pattern_arg.search(arg_text, template_file_end)
      while m_arg != None :
         if m_arg.group(1) != None :
            arg = m_arg.group(1).split(separator)
            if len( arg ) == 1 :
               comment = arg[0].strip()
               if comment not in template_data :
                  msg  = 'xrst_template comment =\n'
                  msg += comment + '\n'
                  msg += f'does not appear in {template_file}'
                  xrst.system_exit(
                     msg,
                     file_name = page_file,
                     page_name = page_name,
                     m_obj     = m_arg,
                     data      = arg_text,
                  )
         m_arg = pattern_arg.search(arg_text, m_arg.end() )
      #
      # replace_list
      replace_list = list()
      m_arg        = pattern_arg.search(arg_text, template_file_end)
      while m_arg != None :
         if m_arg.group(1) != None :
            arg = m_arg.group(1).split(separator)
            if len( arg ) > 2 :
               msg  = f'xrst_template separator = {separator}\n'
               msg += 'separator appears more that once in a line '
               msg += 'below the template_file line.'
               xrst.system_exit(
                  msg,
                  file_name = page_file,
                  page_name = page_name,
                  m_obj     = m_arg,
                  data      = arg_text,
               )
            if len( arg ) == 2 :
               #
               # replace_list
               match    = arg[0].strip()
               replace  = arg[1].strip()
               line     = m_arg.group(2)
               replace_list.append( (match, replace, line) )
         #
         # m_arg
         m_arg = pattern_arg.search(arg_text, m_arg.end() )
      #
      # template_expansion
      template_expansion = template_data
      for match, replace, line in replace_list :
         #
         index = template_expansion.find(match)
         if index == -1 :
            msg  = f'template_command: match = {match}'
            msg += 'This match did not appear in the template file'
            xrst.system_exit(msg,
               file_name = page_file,
               page_name = page_name,
               line      = line
            )
         template_expansion = template_expansion.replace(match, replace)
      #
      # template_expansion
      # no newlines in before of after so that add_line_numbers works properly
      line    = m_template.group(3).strip()
      before  = '@{xrst_template_begin@'
      before += template_file + '@'
      before += line + '@}@'
      after   = '@{xrst_template_end}@'
      assert xrst.pattern['template_begin'].match(before) != None
      assert xrst.pattern['template_end'].match(after) != None
      template_expansion = before + template_expansion + after
      #
      # template_expansion
      template_expansion = xrst.add_line_numbers(template_expansion, page_file)
      #
      # template_expansion
      # Now that line numbers in template expansion are correct,
      # add a newline at the beginning so commands that must start with newline
      # can appear in the first line of the template file.
      index  = template_expansion.find('@}@')
      before = template_expansion[ : index + 3 ]
      after  = template_expansion[ index + 3 : ]
      template_expansion = before + '\n' + after
      #
      # template_expansion
      for cmd in [
         'begin',
         'end',
         'comment_ch',
         'indent',
         'spell',
         'template'
      ] :
         pattern = r'[^\\]{xrst_' + cmd + r'[^a-zA-Z_]'
         m_obj   = re.search(pattern, template_expansion)
         if m_obj != None :
            msg  = f'found {cmd} command in template expansion'
            xrst.system_exit(msg,
               file_name = page_file,
               page_name = page_name,
               m_obj     = m_obj,
               data      = template_expansion
            )
      #
      # data_done, data_out
      data_done = data_out[: m_template.start() + 1] + template_expansion
      data_out  = data_done + data_out[m_template.end() :]
      #
      # m_template
      m_template  = pattern_template.search(data_out , len(data_done) )
   #
   # suspend_command
   # need to run this for suspends in the template expansions
   data_out = xrst.suspend_command(data_out, page_file, page_name)
   #
   # BEGIN_RETURN
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import sys
import os
import re
import xrst
#
# {xrst_begin system_exit dev}
# {xrst_spell
#     msg
# }
# {xrst_comment_ch #}
#
# Form error message and exit
# ###########################
#
# msg
# ***
# Reason for aborting xrst
#
# file_name
# *********
# is the name of the file that contains the begin command for this page.
# This is different from the current input file if we are processing
# a template expansion.
#
# page_name
# *********
# name of page that the error appeared in
#
# m_obj
# *****
# The error was detected in the values returned by this match object.
#
# data
# ****
# is the data that was searched to get the match object m_obj.
# If the error possibly occurred in a template expansion, you must include
# the entire expansion in the data.
#
# line
# ****
# is the line number in the current input file where the error
# was detected.
#
# Prototype
# *********
# {xrst_literal
#  # BEGIN_DEF
#  # END_DEF
#  }
#
# {xrst_end system_exit}
#
# BEGIN_DEF
def system_exit(
   msg, file_name=None, page_name=None, m_obj=None, data=None, line=None
) :
   assert type(msg)       == str
   assert type(file_name) == str or file_name == None
   assert type(page_name) == str or page_name == None
   assert type(line)  in [ int, str ] or line == None
   if m_obj != None :
      assert file_name != None
      assert data      != None
   # END_DEF
   #
   # extra
   project_directory = os.getcwd()
   extra          = f'\nproject_directory = {project_directory}\n'
   #
   # page_line, template_file, template_line
   page_line     = line
   template_file = None
   template_line = None
   if m_obj :
      page_line, template_file, template_line = xrst.file_line(m_obj, data)
   #
   # extra
   if page_name != None :
      extra += f'page = {page_name}\n'
   if file_name != None :
      extra += f'file = {file_name}\n'
   if page_line != None :
      extra += f'line = {page_line}\n'
   if template_file != None :
      extra += f'template_file = {template_file}\n'
      extra += f'template_line = {template_line}\n'
   #
   # breakpoint()
   sys.exit('xrst: Error\n' + msg + extra)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import xrst
import re
#
# {xrst_begin start_end_file dev}
# {xrst_spell
#     cmd
# }
# {xrst_comment_ch #}
#
# Convert literal command start, end from text to line numbers
# ############################################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# page_file
# *********
# is the name of the file that contains the begin command for this page.
# This is different from the current input file if we are processing
# a template expansion.
#
# page_name
# *********
# is the name of the page where the xrst_literal command appears.
#
# input_file
# **********
# is the name of the file where the xrst_literal command appears.
# This is different for *page_file* when the command appears in a
# template expansion.
#
# display_file
# ************
# is the name of the file that we are displaying. If it is not the same as
# input_file, then it must have appeared in the xrst_literal command.
#
# cmd_line
# ********
# If input_file is equal to display_file, the lines of the file
# between line numbers cmd_line[0] and cmd_line[1] inclusive
# are in the xrst_literal command and are excluded from the search.
#
# start_after
# ***********
# is the starting text. There must be one and only one copy of this text in the
# file (not counting the excluded text). This text has no newlines and cannot
# be empty.  If not, an the error is reported and the program stops.
#
# end_before
# **********
# is the stopping text. There must be one and only one copy of this text in the
# file (not counting the excluded text). This text has no newlines and cannot
# be empty.  Furthermore, the stopping text must come after the end of the
# starting text. If not, an the error is reported and the program stops.
#
# start_line
# **********
# is the line number where start_after appears.
#
# end_line
# ********
# is the line number where end_before appears.
#
# m_start
# *******
# is a match object corresponding to the location of start_after.
# It is only used for reporting errors.
#
# m_end
# *****
# is a match object corresponding to the location of end_before.
# It is only used for reporting errors.
#
# m_data
# ******
# is the data for the entire page, including template expansion.
# It corresponds to *m_start* , *m_end* and is only used for reporting errors.
#
# {xrst_end start_end_file}
# BEGIN_DEF
def start_end_file(
   page_file,
   page_name,
   input_file,
   display_file,
   cmd_line,
   start_after,
   end_before,
   m_start,
   m_end,
   m_data,
) :
   assert type(page_file) == str
   assert type(page_name) == str
   assert type(input_file) == str
   assert type(display_file) == str
   assert type(cmd_line[0]) == int
   assert type(cmd_line[1]) == int
   assert cmd_line[0] <= cmd_line[1]
   assert type(start_after) == str
   assert type(end_before) == str
   assert type(m_start) == re.Match
   assert type(m_end) == re.Match
   assert type(m_data) == str
   # END_DEF
   # ------------------------------------------------------------------------
   # exclude_line
   if input_file == display_file :
      exclude_line = cmd_line
   else :
      exclude_line = (0, 0)
   #
   # msg
   msg  = f'in literal command:'
   #
   if start_after == '' :
      msg += ' start_after is empty'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_start, data = m_data
      )
   if end_before == '' :
      msg += ' end_before is empty'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_end, data = m_data
      )
   if 0 <= start_after.find('\n') :
      msg += ' a newline appears in start_after'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_start, data = m_data
      )
   if 0 <= end_before.find('\n') :
      msg += ' a newline appears in end_before'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_end, data = m_data
      )
   #
   # data
   file_obj  = open(display_file, 'r')
   data      = file_obj.read()
   file_obj.close()
   #
   # start_line
   start_index = data.find(start_after)
   count = 0
   while 0 <= start_index :
      line = data[: start_index].count('\n') + 1
      if  line < exclude_line[0] or exclude_line[1] < line :
         start_line = line
         count      = count + 1
      start_index = data.find(start_after, start_index + len(start_after) )
   if count != 1 :
      msg += f'\nstart_after   =  {start_after}'
      msg += f'\ndisplay_file  =  {display_file}'
      msg += f'\nfound {count} matches expected 1'
      if input_file == display_file :
         msg += ' not counting the literal command'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_start, data = m_data
      )
   #
   # end_line
   stop_index = data.find(end_before)
   count = 0
   while 0 <= stop_index :
      line = data[: stop_index].count('\n') + 1
      if  line < exclude_line[0] or exclude_line[1] < line :
         end_line = line
         count     = count + 1
      stop_index = data.find(end_before, stop_index + len(end_before) )
   if count != 1 :
      msg += f'\nend_before   =  {end_before}'
      msg += f'\ndisplay_file =  {display_file}'
      msg += f'\nfound {count} matches expected 1'
      if input_file == display_file :
         msg += ' not counting the literal command'
      xrst.system_exit(msg, file_name=page_file, page_name=page_name,
         m_obj = m_end, data = m_data
      )
   # ------------------------------------------------------------------------
   # BEGIN_RETURN
   #
   assert type(start_line) == int
   assert type(end_line) == int
   return start_line, end_line
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin code_cmd user}

Code Command
############

Syntax
******
- ``\{xrst_code`` *language* ``}``
- ``\{xrst_code}``

Purpose
*******
A code block, directly below in the current input file, begins (ends) with
a code command that contains *language* (not containing *language*).

code-block
**********
This command is similar to the following sphinx directive

| |tab| .. code-block:: *language*
|
| |tab| |tab| Source code in the specified language


The xrst code command has the following difference:

#. Other characters on the same line as the code commands
   are not included in the rst output or the displayed source code.
   One can use these characters to end and begin comments so that the
   code also executes.
#. The source code does not need to be indented. This is especially useful
   with languages like python and rst where the indentation is used to
   begin and end sections of code.
#. The source code does not need to be surrounded by empty lines.

Requirements
************
Each code section ends with
a line containing the second version of the command; i.e., ``\{xrst_code}``.
Hence there must be an even number of code commands.

language
********
A *language* is a non-empty sequence of lower case letters.
It determines the language for highlighting the code block.

Rest of Line
************
Other characters on the same line as a code commands
are not included in the xrst output.
This enables one to begin or end a comment block
without having the comment characters in the xrst output.

Spell Checking
**************
Code blocks as usually small and
spell checking is done for these code blocks.
You can turn off this spell checking by putting
:ref:`spell_cmd@spell_off` before and :ref:`spell_cmd@spell_on` after
a code block.
Spell checking is not done for code blocks included using the
:ref:`literal command<literal_cmd-name>` .

Example
*******
:ref:`code_example-name`

{xrst_end code_cmd}
"""
# ----------------------------------------------------------------------------
import xrst
import os
# {xrst_begin code_cmd_dev dev}
# {xrst_comment_ch #}
#
# Process the xrst code commands for a page
# #########################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for the page before the
# :ref:`code commands <code_cmd-name>` have been processed.
#
# file_name
# *********
# is the name of the file that this data comes from. This is only used
# for error reporting.
#
# page_name
# *********
# is the name of the page that this data is in. This is only used
# for error reporting.
#
# rst2project_dir
# ***************
# is a relative path from the :ref:`config_file@directory@rst_directory`
# to the :ref:`config_file@directory@project_directory` .
#
# data_out
# ********
# is a copy of data_in with the xrst code commands replaced by a corresponding
# sphinx command.
#
# {xrst_end code_cmd_dev}
# BEGIN_DEF
def code_command(data_in, file_name, page_name, rst2project_dir) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   assert type(rst2project_dir) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_begin
   m_begin = xrst.pattern['code'].search(data_out)
   #
   if m_begin == None :
      return data_out
   #
   while m_begin != None :
      #
      # m_end
      start = m_begin.end()
      m_end = xrst.pattern['code'].search(data_out, start)
      #
      # language
      language  = m_begin.group(4).strip()
      if language == '' :
         msg = 'missing language in first command of a code block pair'
         xrst.system_exit(msg,
            file_name=file_name,
            page_name=page_name,
            m_obj=m_begin,
            data=data_out
         )
      for ch in language :
         if ch < 'a' or 'z' < ch :
            msg = 'code block language character not in a-z.'
            xrst.system_exit(msg,
               file_name=file_name,
               page_name=page_name,
               m_obj=m_begin,
               data=data_out
            )
      #
      if m_end == None :
         msg = 'Start code command does not have a corresponding stop'
         xrst.system_exit(msg,
            file_name=file_name,
            page_name=page_name,
            m_obj=m_begin,
            data=data_out
         )
      if m_end.group(4).strip() != '' :
         msg ='Stop code command has a non-empty language argument'
         xrst.system_exit(msg,
            file_name=file_name,
            page_name=page_name,
            m_obj=m_end,
            data=data_out
         )
      #
      # language
      # fix cases that pygments has trouble with ?
      if language == 'hpp' :
         language = 'cpp'
      if language == 'm' :
         language = 'matlab'
      #
      # data_before, data_between,  data_after
      data_before  = data_out[: m_begin.start()]
      data_between = data_out[ m_begin.start() : m_end.end() ]
      data_after   = data_out[m_end.end() : ]
      #
      # indent
      indent = xrst.auto_indent(data_between, file_name, page_name)
      #
      # start_line, stop_line
      assert data_out[m_begin.end()] == '\n'
      assert data_out[m_end.end()] == '\n'
      start_line = int(m_begin.group(5)) + 1
      stop_line  = int(m_end.group(5)) - 1
      #
      # cmd
      file_path = os.path.join(rst2project_dir, file_name)
      command   = indent + f'.. literalinclude:: {file_path}\n'
      command  += indent + 3 * ' ' + f':lines: {start_line}-{stop_line}\n'
      command  += indent + 3 * ' ' + f':language: {language}\n'
      command  = '\n' + command + '\n'
      assert data_after.startswith('\n')
      if not data_before.strip(' ').endswith('\n') :
         command = '\n' + command
      #
      # data_left, data_after
      data_left  = data_before + command
      data_out   = data_left + data_after
      #
      # m_begin
      m_begin = xrst.pattern['code'].search(data_out, len(data_left) )
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin comment_cmd user}

Comment Command
###############

Syntax
******
``\{xrst_comment`` *text* ``}``

Purpose
*******
This command adds *text* to the xrst input file that
does not appear in the rst output file.

Example
*******
:ref:`comment_example-name`

{xrst_end comment_cmd}
"""
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern
pattern = re.compile(
   r'([^\n]*[^\\]){xrst_comment[^a-z_][^}]*}([^\n]*\n)'
)
# {xrst_begin comment_cmd_dev dev}
# {xrst_comment_ch #}
#
# Remove all comment commands
# ###########################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page.
#
# data_out
# ********
# The return data_out is a copy of data_in except that the comment
# commands have been removed.
#
# {xrst_end comment_cmd_dev}
# BEGIN_DEF
def comment_command(data_in) :
   assert type(data_in) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_obj
   m_obj = pattern.search(data_out)
   while m_obj :
      # data_before, data_after
      data_before = data_out[: m_obj.start()]
      data_after  = data_out[m_obj.end() :]
      #
      # text_before, text_after
      text_before = m_obj.group(1).rstrip(' \t')
      text_after  = m_obj.group(2).lstrip(' \t')
      #
      # data_before, text_before
      if text_before.endswith('\n') :
         data_before += text_before
         text_before  = ''
      #
      # data_after, text_after
      if text_before == '' and text_after.startswith('@xrst_line ') :
         text_after = ''
      #
      # data_out
      if text_before != '' and text_after != '' :
         other_text = text_before + ' ' + text_after
      else :
         other_text = text_before + text_after
      data_out   = data_before + other_text +  data_after
      #
      # m_obj
      m_obj = pattern.search(data_out, m_obj.start())
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
# PAGE_NAME_PATTERN = [A-Za-z0-9._-]+
# use git grep PAGE_NAME_PATTERN to get all occurrences of this pattern
#
# {xrst_begin check_page_name dev}
# {xrst_spell
#     genindex
#     underbar
# }
# {xrst_comment_ch #}
#
# Check the rules for a page name
# ###############################
#
# page_name
# *********
# The page_name appears in *m_obj* in one of the following ways
#
# #.  \{xrst_begin_parent page_name user}
# #.  \{xrst_begin page_name user}
# #.  \{xrst_end page_name}
#
# A valid page name must satisfy the following conditions:
#
# #.  The valid characters in a page name are [A-Z], [a-z], [0-9],
#     dash, period and underbar.
# #.  A page name cannot begin with ``xrst_`` .
# #.  A page name cannot be ``index`` or ``genindex`` .
#
# If *page_name* does not follow
# these rules, a message is printed and the program exits.
#
# file_name
# *********
# is the name of the original input file that data appears in
# (used for error reporting).
#
# m_obj
# *****
# is the match object corresponding to *page_name*
#
# data
# ****
# is that data that was searched to get the match object.
#
# Prototype
# *********
# {xrst_literal
#  # BEGIN_DEF
#  # END_DEF
#  }
#
# {xrst_end check_page_name}
#
# BEGIN_DEF
def check_page_name(page_name, file_name, m_obj, data) :
   assert type(page_name) == str
   assert type(file_name) == str
   assert m_obj
   assert type(data) == str
   # END_DEF
   #
   m_page_name = re.search('[-._A-Za-z0-9]+', page_name)
   if m_page_name.group(0) != page_name :
      msg  = f'begin command: page_name = "{page_name}"'
      msg += '\nIt must be non-empty and only contain the following'
      msg += ' characters: -, ., _, A-Z, a-z, 0-9'
      xrst.system_exit(msg,
         file_name=file_name, m_obj=m_obj, data=data
      )
   if page_name.startswith('xrst_') :
      msg = 'page_name cannot start with xrst_'
      xrst.system_exit(msg,
         file_name=file_name, m_obj=m_obj, data=data
      )
   if page_name == 'index' or page_name == 'genindex' :
      msg = 'page_name cannot be index or genindex'
      xrst.system_exit(msg,
         file_name=file_name, m_obj=m_obj, data=data
      )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
pattern_line_number     = re.compile(r'\n[ \t]*@xrst_line [0-9]+@')
pattern_newline_3       = re.compile(r'(\n[ \t]*){2,}\n')
pattern_white_at_eol    = re.compile(r'[ \t]*\n')
#
# PAGE_NAME_PATTERN = [A-Za-z0-9._-]+
# use git grep PAGE_NAME_PATTERN to get all occurrences of this pattern
pattern_ref_page_name_1 = re.compile(r':ref:`([A-Za-z0-9._-]+)-name`')
pattern_ref_page_name_2 = re.compile(
                           r':ref:`([^`<]*)<([A-Za-z0-9._-]+)-name>`'
)
pattern_any_command     = re.compile(r'[^\\]{xrst_([a-zA-Z_]*)')
# ----------------------------------------------------------------------------
# {xrst_begin temporary_file dev}
# {xrst_spell
#     tuple
# }
# {xrst_comment_ch #}
#
# Write the temporary RST file for a page
# #######################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# page_source
# ***********
# If *page_source* is true and *target* is html,
# a link to the xrst source code is included at the
# top of each page.
# If *page_source* is true and *target* is tex,
# the location of the xrst source code is reported at the
# top of each page.
#
# target
# ******
# If the :ref:`run_xrst@target` command line argument.
#
# pseudo_heading
# **************
# is the :ref:`process_headings@pseudo_heading` for this page.
#
# file_in
# *******
# is the name of the xrst input file for this page.
#
# tmp_dir
# *******
# is the directory where the output file will be saved.
#
# page_name
# *********
# is the name of this page.
#
# file_out
# ********
# The temporary file written by the routine, *file_out* , is
# tmp_dir/page_name.rst.

# data_in
# *******
# is the data for this page with all the xrst commands converted to
# their sphinx RST values, except the \\n\{xrst\@before_title} command.
# The following is added to this data before writing it to the output file:
#
#  #. The preamble is included at the beginning.
#  #. If *target* is ``html`` :
#
#     #. The *page_name* ``-name`` label is added next.
#     #. The pseudo heading is added next.
#     #. The name of the input file *file_in* is added next.
#
#  #. If *target* is ``tex`` :
#     All cross references of the form :ref:\` *page_name* -name \` ,
#     are changed to :ref:\` *page_name* < *page_name* -title >\` .
#     Note that the page name will be added to the title when
#     :ref:`add_before_title-name` is called during
#     :ref:`table_of_contents-name` .
#
#  #. Check for an xrst command that was not recognized.
#  #. Any sequence of more than 2 lines
#     with only tabs or space are converted to 2 empty lines.
#  #. Empty lines at the end are removed
#  #. The xrst_line_number entries are removed.
#  #. The xrst_template_begin and xrst_template_end markers are removed.
#  #. The text ``\\{xrst_`` is replaced by ``\{xrst_`` .
#
# rst2xrst_list
# *************
# This is the value returned by ``temporary_file`` .
# For each *index*, *rst2xrst_list* [ *index* ] is the a tuple; see
# :ref:`remove_line_numbers@rst2xrst_list` .
#
# {xrst_end temporary_file}
# BEGIN_DEF
# rst2xrst_list =
def temporary_file(
   page_source,
   target,
   pseudo_heading,
   file_in,
   tmp_dir,
   page_name,
   data_in,
) :
   assert target == 'html' or target == 'tex'
   assert type(page_source) == bool
   assert type(pseudo_heading) == str
   assert type(file_in) == str
   assert type(page_name) == str
   assert type(data_in) == str
   # END_DEF
   #
   # punctuation
   # Headings uses repeated copies of one of these characters
   punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
   #
   # label
   # label that displays page name (which is text in pseudo heading)
   label = f'.. _{page_name}-name:\n\n'
   #
   # data_out
   if target == 'html' :
      before   = label
      before  += pseudo_heading
      if page_source :
         if page_name.endswith('.rst') :
            destination = f'_sources/{page_name}.txt'
         else :
            destination = f'_sources/{page_name}.rst.txt'
         raw_html  = '.. raw:: html\n\n' + 3 * ' '
         raw_html += f'<a href="{destination}">View page source</a>\n\n'
         before   += raw_html
      data_out  = before + data_in
   else :
      if page_source :
         new_text   = f'**xrst input file** : {file_in}\n\n'
         #
         index      = data_in.find('\n{xrst@before_title}')
         first_line = index + len('\n{xrst@before_title}')
         assert data_in[first_line] == '\n'
         second_line = data_in.find('\n', first_line + 1)
         third_line  = data_in.find('\n', second_line + 1)
         fourth_line = data_in.find('\n', third_line + 1)
         #
         line     = data_in[first_line + 1 : second_line]
         overline = line[0] * len(line) == line and line[0] in punctuation
         if overline :
            text_index = fourth_line + 1
         else :
            text_index = third_line + 1
         data_before = data_in[: text_index]
         data_after  = data_in[text_index :]
         data_out    = data_before + new_text + data_after
      else :
         data_out = data_in
      #
      data_out   = pattern_ref_page_name_1.sub(
         r':ref:`\1<\1-title>`' , data_out
      )
      data_out   = pattern_ref_page_name_2.sub(
         r':ref:`\1<\2-title>`' , data_out
      )
   #
   # m_obj
   m_obj = pattern_any_command.search( data_out )
   while m_obj != None :
      ok = m_obj.group(0) == '@{xrst_template_begin'
      ok = ok or m_obj.group(0) == '@{xrst_template_end'
      if not ok :
         cmd  = m_obj.group(1)
         msg  = f'This xrst {cmd} command was not recognized.\n'
         msg += 'Use \\{xrst_' + cmd + ' '
         msg += 'if this is not intented to be an xrst command.'
         xrst.system_exit(msg,
            file_name=file_in,
            page_name=page_name,
            m_obj=m_obj,
            data=data_out
         )
      m_obj = pattern_any_command.search( data_out , m_obj.end() )
   #
   # data_out
   # Convert three or more sequential emtpty lines to two empty lines.
   data_out = pattern_line_number.sub('\n', data_out)
   data_out = pattern_newline_3.sub('\n\n', data_out)
   #
   # data_out
   # remove empty lines at the end
   while data_out[-2 :] == '\n\n' :
      data_out = data_out[: -1]
   #
   # data_out
   # The last step removing line numbers. This is done last for two reasons:
   # 1. So mapping from output to input line number is correct.
   # 2. We are no longer able to give line numbers for errors after this.
   data_out, rst2xrst_list = xrst.remove_line_numbers(data_out)
   #
   # data_out
   # Need the template_begin and template_end while removing line numbers
   # so that rst2xrst_list contains the template file information.
   data_out = xrst.pattern['template_begin'].sub('', data_out)
   data_out = xrst.pattern['template_end'].sub('', data_out)
   #
   # data_out
   data_out = data_out.replace( r'\{xrst_', '{xrst_' )
   #
   # data_out
   # remove white space at end of each line
   data_out = pattern_white_at_eol.sub(r'\n', data_out)
   #
   # file_out
   if page_name.endswith('.rst') :
      file_out = tmp_dir + '/' + page_name
   else :
      file_out = tmp_dir + '/' + page_name + '.rst'
   file_obj = open(file_out, 'w')
   file_obj.write(data_out)
   file_obj.close()
   #
   # BEGIN_RETURN
   #
   assert type(rst2xrst_list) == list
   assert type( rst2xrst_list[0] ) == tuple
   assert len( rst2xrst_list[0] ) == 2 or len( rst2xrst_list[0] ) == 4
   assert type( rst2xrst_list[0][0] ) == int
   assert type( rst2xrst_list[0][1] ) == int
   return rst2xrst_list
   # END_RETURN



#! /usr/bin/env python
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import sys
import os
import re
import subprocess
# ----------------------------------------------------------------------------
pattern_return_newline    = re.compile( r'\r\n' )
pattern_include_backslash = re.compile( r'\.\. literalinclude:: .*\\' )
# ----------------------------------------------------------------------------
def get_index_page_name() :
   file_obj  = open('.readthedocs.yaml', 'r')
   file_data = file_obj.read()
   pattern   = r'\n *--index_page_name *([-._a-zA-Z0-9]*)'
   m_obj     = re.search(pattern, file_data)
   file_obj.close()
   return m_obj.group(1)
# ----------------------------------------------------------------------------
def get_rst_directory() :
   file_obj  = open('xrst.toml', 'r')
   file_data = file_obj.read()
   pattern   = r"\nrst_directory *= *'([^']*)'"
   m_obj     = re.search(pattern, file_data)
   file_obj.close()
   return m_obj.group(1)
# ----------------------------------------------------------------------------
def run_xrst(
   test_installed_version,
   test_external_links,
   suppress_spell_warnings,
) :
   index_page_name = get_index_page_name()
   #
   # python_executable
   python_executable = sys.executable
   #
   # command
   if test_installed_version :
      if not os.path.exists('build') :
         os.mkdir('build')
      os.chdir('build')
      command  = [ 'xrst', '--config_file', '../xrst.toml' ]
   else :
      command  = [ python_executable, '-m', 'xrst' ]
      command += [ '--config_file', 'xrst.toml' ]
      # sphinx_rtd_theme is in the tox.ini file as a dependency
      command += [ '--html_theme',      'sphinx_rtd_theme' ]
   if test_external_links :
      command += [ '--external_links' ]
   if suppress_spell_warnings :
      command += [ '--suppress_spell_warnings' ]
   #
   # command
   # This command should have the same group_list as the last group_list in
   # bin/check_xrst.sh which is used to keep the test_rst up to date.
   command += [
      '--local_toc',
      '--rst_only',
      '--index_page_name', index_page_name,
      '--group_list',      'default', 'user', 'dev',
   ]
   result = subprocess.run(command)
   assert result.returncode == 0
   #
   if test_installed_version :
      os.chdir('..')
# ----------------------------------------------------------------------------
def run_test(
   test_installed_version,
   test_external_links,
   suppress_spell_errors,
) :
   #
   # run_xrst
   run_xrst(
      test_installed_version,
      test_external_links,
      suppress_spell_errors
   )
   #
   # rst_list
   rst_directory = get_rst_directory()
   rst_list      = list()
   for entry in os.listdir(rst_directory) :
      if entry.endswith('.rst') :
         rst_list.append(entry)
   rst_list = sorted(rst_list)
   #
   # check_list
   check_list = sorted( os.listdir('test_rst') )
   #
   # rst_index, check_index
   rst_index   = 0
   check_index = 0
   while rst_index < len(rst_list) and check_index < len(check_list) :
      #
      # rst_name, check_name
      rst_name   = rst_list[rst_index]
      check_name = check_list[check_index]
      if rst_name < check_name :
         rst_index += 1
         assert False, f'{rst_name} is in {rst_directory} but not test_rst'
      elif check_name < rst_name :
         check_index += 1
         assert False, f'{check_name} is in test_rst but not {rst_directory}'
      else :
         #
         # rst_index, check_index
         rst_index     += 1
         check_index   += 1
         #
         # rst_file_data, check_file_data
         rst_file      = f'{rst_directory}/{rst_name}'
         check_file    = f'test_rst/{check_name}'
         #
         rst_file_obj   = open(rst_file, 'r')
         check_file_obj = open(check_file, 'r')
         #
         rst_data      = rst_file_obj.read()
         check_data    = check_file_obj.read()
         #
         rst_file_obj.close()
         check_file_obj.close()
         #
         # rst_data
         # dos2unix: relpace \r\n by \n
         replace  = r'\n'
         rst_data = pattern_return_newline.sub(replace, rst_data)
         #
         # rst_data
         # relpace \ in file names by /
         m_obj = pattern_include_backslash.search(rst_data)
         while m_obj != None :
            start      = m_obj.start()
            end        = m_obj.end()
            text       = m_obj.group(0)
            replace    = text.replace( '\\' , '/' )
            rst_data   = check_data[:start] + replace + rst_data[end:]
            m_obj      = pattern_include_backslash.search(rst_data)
         #
         if rst_data == check_data :
            print( f'{rst_name}: OK' )
         else :
            msg = f'{rst_file} is different from {check_file}'
            assert False, msg
# ----------------------------------------------------------------------------
def test_rst(
   test_installed_version  = False,
   test_external_links     = True,
   suppress_spell_warnings = False,
) :
   print( f'test_installed_version   = {test_installed_version}' )
   print( f'test_external_links      = {test_external_links}' )
   print( f'suppress_spell_warnings  = {suppress_spell_warnings}' )
   if not os.path.exists('xrst.toml') :
      assert False, 'test_rst.py: can not find xrst.toml in working directory'
   else :
      run_test(
         test_installed_version  ,
         test_external_links     ,
         suppress_spell_warnings ,
      )
# ----------------------------------------------------------------------------
if __name__ == '__main__' :
   #
   # usage
   usage   = 'usage: pytest/test_rst.py test_installed_version '
   usage  += 'test_external_links suppress_spell_warnings\n'
   usage  += 'where all the arguments are True or False'
   #
   # test_installed_version
   program = sys.argv[0]
   if program != 'pytest/test_rst.py' or len(sys.argv) != 4 :
      sys.exit(usage)
   for i in range(1, 4) :
      if sys.argv[i] not in [ 'True', 'False' ] :
         sys.exit(usage)
   test_installed_version  = sys.argv[1] == 'True'
   test_external_links     = sys.argv[2] == 'True'
   suppress_spell_warnings = sys.argv[3] == 'True'
   #
   test_rst(
         test_installed_version,
         test_external_links,
         suppress_spell_warnings
   )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import tomli
import xrst
# ----------------------------------------------------------------------------
# {xrst_begin rename_group dev}
# {xrst_comment_ch #}
#
# Rename a Subset of a Group
# ##########################
#
# tmp_dir
# *******
# is the directory where spell.toml is located
#
# old_group_name
# **************
# is the old name that we are replacing in the xrst begin commands.
#
# new_group_name
# **************
# is the new name that we are using in the xrst begin commands.
#
# spell.toml
# **********
# see :ref:`replace_spell@spell.toml` .
#
# Prototype
# *********
# {xrst_literal
#  # BEGIN_DEF
#  # END_DEF
#  }
#
# {xrst_end rename_group}
#
# BEGIN_DEF
def rename_group(tmp_dir, old_group_name, new_group_name) :
   assert type(tmp_dir) == str
   assert type(old_group_name) == str
   assert type(new_group_name) == str
   # END_DEF
   #
   # spell_toml
   file_obj  = open( f'{tmp_dir}/spell.toml' )
   file_data = file_obj.read()
   spell_toml = tomli.loads(file_data)
   #
   # file_name
   for file_name in spell_toml :
      #
      # page_list
      page_list = list()
      for page_name in spell_toml[file_name] :
         pair = (page_name , spell_toml[file_name][page_name] )
         page_list.append( pair )
      #
      # page_list
      order_fun    = lambda page_pair : page_pair[1]['begin_line']
      page_list = sorted(page_list, key = order_fun )
      #
      # data_in
      file_obj = open(file_name, 'r')
      data_in  = file_obj.read()
      file_obj.close()
      #
      # data_out
      data_out      = ''
      data_in_index = 0
      #
      # page_name, page_info
      for page_name, page_info in page_list :
         #
         # m_begin
         m_begin = xrst.pattern['begin'].search(data_in, data_in_index)
         while m_begin.group(3) != page_name :
            m_begin = xrst.pattern['begin'].search(data_copy, m_begin.end())
         #
         # before_begin
         before_begin = m_begin.group(1)
         #
         # begin_type
         begin_type = m_begin.group(2)
         #
         # this_page_name
         this_page_name = m_begin.group(3)
         assert this_page_name == page_name
         #
         # this_group_name
         this_group_name = m_begin.group(4).strip()
         if this_group_name == '' :
            this_group_name = 'default'
         assert this_group_name == old_group_name
         #
         # data_before, data_after
         data_before = data_in[data_in_index : m_begin.start() ]
         #
         # begin_command
         begin_command = before_begin + '{' + \
               f'xrst_{begin_type} {page_name} {new_group_name}' + \
         '}'
         #
         # data_out
         data_out += data_before + begin_command
         #
         # data_in_index
         data_in_index = m_begin.end()
      #
      # data_out
      data_out += data_in[data_in_index :]
      #
      # file_name
      file_obj = open(file_name, 'w')
      file_obj.write( data_out )
      file_obj.close()
   #
   return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin begin_cmd user}
{xrst_spell
   genindex
   underbar
}

Begin and End Commands
######################

Syntax
******
- ``\{xrst_begin_parent`` *page_name* *group_name* ``}``
- ``\{xrst_begin``        *page_name* *group_name* ``}``
- ``\{xrst_end``          *page_name* ``}``

Page
****
The start (end) of a page of the input file is indicated by a
begin (end) command.

Page File
*********
The file containing the begin and end commands for a page
is called its page file.
During the processing of :ref:`template commands<template_cmd-name>`
the page file is different from the current input file.

page_name
*********
A *page_name* must satisfy the following conditions:

#. It must be a non-empty sequence of the following characters:
   dash ``-``, period ``.``, underbar ``_``, the letters A-Z, letters a-z,
   and decimal digits 0-9.
#. The page name can not be ``index`` or ``genindex`` ,
   and it can not begin with the characters ``xrst_``.
#. The lower case version of two page names cannot be equal.

A link is included in the index under the page name to the page.
The page name is also added to the html keyword meta data.

group_name
**********
The *group_name* can be empty or a sequence of the letters a-z.
This is the group that this page belongs to; see
:ref:`run_xrst@group_list`.

Default Group
=============
The default value for *group_name* is ``default``; i.e.,
if *group_name* is the empty string, this page is part of the default group.

Output File
***********
The output file corresponding to *page_name* is

   *rst_directory*\ /\ *page_name*\ /``.rst``

see :ref:`config_file@directory@rst_directory` .

Parent Page
***********
The following conditions hold for each *group_name*:

#. There can be at most one begin parent command in an input file.
#. If there is a begin parent command, it must be the first begin command
   in the file and there must be other pages in the file.
#. The other pages are children of the parent page.
#. The parent page is a child
   of the page that included this file using a
   :ref:`toc command<toc_cmd-name>`.
#. If there is no begin parent command in an input file,
   all the pages in the file are children
   of the page that included this file using a
   :ref:`toc command<toc_cmd-name>`.

Note that there can be more than one begin parent command in a file if
they have different group names. Also note that pages are only children
of pages that have the same group name.

{xrst_end begin_cmd}
"""
# ---------------------------------------------------------------------------
import xrst
import re
pattern_group_name    = re.compile( r'[^ \t]+' )
pattern_group_valid   = re.compile( r'[a-z]+' )
pattern_template_file = re.compile(
   r'([^\\]){xrst_template[^\n}]*\n([^\n}]*)@xrst_line [0-9]+@\n'
)
# ---------------------------------------------------------------------------
# {xrst_begin get_file_info dev}
# {xrst_spell
#     len
# }
# {xrst_comment_ch #}
#
# Get information for all pages in a file
# #######################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# all_page_info
# *************
# a list of with information for pages that came before this file.
# For each list index, all_page_info[index] is a dict and
# all_page_info[index]['page_name'] is an str
# containing the name of a page that came before this file.
# This includes pages for all the groups that came before this group.
#
# group_name
# **********
# We are only retrieving information for pages in this group.
# (This is non-empty because default is used for the empty group name.)
#
# parent_file
# ***********
# name of the file that included file_in.
#
# file_in
# *******
# is the name of the file we are getting all the information for.
#
# file_page_info
# **************
# The value file_page_info is a list of dict.
# Each dict contains the information
# for one page in this file. We use info below for one element of the list:
#
# info['page_name']
# =================
# is an str containing the name of a page in this file.
#
# info['page_data']
# =================
# is an str containing the data for this page.
# This data has been processed in the following way and order.
#
#  #. Line numbers have been added using :ref:`add_line_numbers-name` .
#     This is the first operation done on a page and other operations
#     assume that line numbers are present. They are removed near the end
#     when the temporary file corresponding to a page is created.
#  #. The page data has been restricted to the text between
#     the end of the begin command and the start of the end command.
#  #. The suspend / resume commands and data between such pairs
#     have been removed; see :ref:`suspend_cmd-name` .
#  #. The indentations for this page have been removed; see
#     :ref:`indent_cmd-name` .
#  #. If a comment character command is present for this page,
#     the command is remove and for each line, the possible
#     comment character and possible space after have been removed.
#
# info['is_parent']
# =================
# is true (false) if this is (is not) the parent page for the other
# pages in this file. The parent page must be the first for this group,
# and hence have index zero in file_info. In addition,
# if there is a parent page, there must be at least one other page;
# i.e., len(file_info) >= 2.
#
# info['is_child']
# ================
# is true (false) if this is (is not) a child of the first page in
# this file.
#
# info['begin_line']
# ==================
# is the line number in *file_in* where this page begins; i.e.,
# the line number where the begin command is located.
#
# info['end_line']
# ================
# is the line number in *file_in* where this page ends; i.e.,
# the line number where the end command is located.
#
# {xrst_end get_file_info}
# BEGIN_DEF
def get_file_info(
      all_page_info,
      group_name,
      parent_file,
      file_in,
) :
   assert type(all_page_info) == list
   if 0 < len(all_page_info) :
      type( all_page_info[0] ) == dict
   assert type(group_name) == str
   assert group_name != ''
   assert type(parent_file) == str or parent_file == None
   assert type(file_in) == str
   # END_DEF
   #
   # file_data
   file_obj   = open(file_in, 'r')
   file_data  = file_obj.read()
   file_obj.close()
   #
   # file_data
   file_data = xrst.add_line_numbers(file_data, file_in)
   #
   # file_page_info
   file_page_info = list()
   #
   # parent_page_name
   parent_page_name = None
   #
   # data_index
   # index to start search for next pattern in file_data
   data_index  = 0
   #
   # found_group_name
   found_group_name = False
   #
   # for each page in this file
   while data_index < len(file_data) :
      #
      # m_begin
      m_begin = xrst.pattern['begin'].search(file_data, data_index)
      #
      # this_group_name
      # This match can't occur in a template expansion.
      if m_begin != None :
         #
         this_group_name = m_begin.group(4)
         m_group         = pattern_group_name.search(this_group_name)
         if m_group == None :
            this_group_name = 'default'
         else :
            this_group_name = m_group.group(0)
            m_group    = pattern_group_valid.search(this_group_name)
            if this_group_name != m_group.group(0) :
               msg = f'"{this_group_name}" is not a valid group name'
               xrst.system_exit(msg,
                  file_name = file_in,
                  m_obj     = m_begin,
                  data      = file_data,
               )
      if m_begin == None :
         if not found_group_name :
            msg  = 'can not find a begin command with \n'
            if group_name == '' :
               msg += 'the empty group name and '
            else :
               msg += f'group_name = {group_name} and '
            msg += f'parent file = {parent_file}'
            xrst.system_exit(msg, file_name=file_in)
         #
         # data_index
         # set so that the page loop for this file terminates
         data_index = len(file_data)
      elif this_group_name != group_name :
         #
         # data_index
         # place to start search for next page
         data_index = m_begin.end()
      else :
         #
         # found_group_name
         found_group_name = True
         #
         # page_name, is_parent
         page_name = m_begin.group(3)
         is_parent = m_begin.group(2) == 'begin_parent'
         #
         # check_page_name
         xrst.check_page_name(
            page_name,
            file_name     = file_in,
            m_obj         = m_begin,
            data          = file_data
         )
         #
         # check if page_name appears multiple times in this file
         for info in file_page_info :
            previous_page_name = info['page_name']
            if page_name.lower() == previous_page_name.lower() :
               msg  = 'Lower case version of two page names are equal.\n'
               msg += f'previous page_name = {previous_page_name}\n'
               msg += f'previous_file      = {file_in}\n'
               xrst.system_exit(msg,
                  file_name      = file_in,
                  page_name      = page_name,
                  m_obj          = m_begin,
                  data           = file_data
               )
         #
         # check if page_name appears in another file
         for info in all_page_info :
            previous_page_name = info['page_name']
            if page_name.lower() == previous_page_name.lower() :
               msg  = 'Lower case version of two page names are equal.\n'
               msg += f'previous_page_name  = "{previous_page_name}"\n'
               msg += 'previous_file        = ' +  info['file_in'] + '\n'
               xrst.system_exit(msg,
                  file_name = file_in  ,
                  page_name = page_name
               );
         #
         # check if parent pages is the first page in this file
         if is_parent :
            if len(file_page_info) != 0 :
               msg  = 'xrst_begin_parent'
               msg += ' is not the first begin command in this file'
               xrst.system_exit(msg,
                  file_name     = file_in,
                  page_name     = page_name,
                  m_obj         = m_begin,
                  data          = file_data
               )
            #
            # parent_page_name
            parent_page_name = page_name
         #
         # is_child
         is_child = (not is_parent) and (parent_page_name != None)
         #
         # data_index
         data_index = m_begin.end()
         #
         # begin_line
         m_line     = xrst.pattern['line'].search(file_data, data_index)
         begin_line = int( m_line.group(1) )
         #
         # m_end
         m_end     = xrst.pattern['end'].search(file_data, data_index)
         #
         if m_end == None :
            msg  = 'Could not find the following text:\n'
            msg += '    {xrst_end ' + page_name + '}'
            xrst.system_exit(
               msg, file_name=file_in, page_name=page_name
            )
         if m_end.group(1) != page_name :
            msg  = 'begin and end page names do not match\n'
            msg += 'begin name = ' + page_name + '\n'
            msg += 'end name   = ' + m_end.group(1)
            xrst.system_exit(msg,
               file_name = file_in,
               m_obj     = m_end,
               data      = file_data
            )
         #
         # end_line
         m_line     = xrst.pattern['line'].search(file_data, m_end.start())
         end_line = int( m_line.group(1) )
         #
         #
         # page_data
         page_start = data_index
         page_end   = m_end.start() + 1
         page_data  = file_data[ page_start : page_end ]
         #
         # page_data
         # order of these operations is important
         page_data = xrst.suspend_command( page_data, file_in, page_name)
         page_data = xrst.indent_command( page_data, file_in, page_name)
         #
         # page_data
         page_data, comment_ch = xrst.comment_ch_command(
            page_data, file_in, page_name
         )
         if comment_ch :
            pattern_ch  = re.compile( r'\n[ \t]*[' + comment_ch + r'] ?' )
            page_data   = pattern_ch.sub(r'\n', page_data)
         #
         # template_list
         template_list = list()
         m_obj         = pattern_template_file.search(page_data)
         while m_obj != None :
            template_list.append( m_obj.group(2).strip() )
            m_obj = pattern_template_file.search( page_data, m_obj.end() )
         #
         # file_page_info
         file_page_info.append( {
            'page_name'      : page_name,
            'page_data'      : page_data,
            'is_parent'      : is_parent,
            'is_child'       : is_child,
            'begin_line'     : begin_line,
            'end_line'       : end_line,
            'template_list'  : template_list,
         } )
         #
         # data_index
         # place to start search for next page
         data_index = m_end.end()
   #
   if parent_page_name != None and len(file_page_info) < 2 :
      msg  = 'begin_parent command appears with '
      if group_name == '' :
         msg += 'the empty group name\n'
      else :
         msg += f'group_name = {group_name}\n'
      msg += 'and this file only has one page with that group name.'
      xrst.system_exit(
         msg, file_name=file_in, page_name=parent_page_name
      )
   #
   # file_data
   file_data = xrst.pattern['begin'].sub('', file_data)
   file_data = xrst.pattern['end'].sub('', file_data)
   #
   # BEGIN_RETURN
   #
   assert type(file_page_info) == list
   assert type(file_page_info[0]) == dict
   return file_page_info
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
pattern_start_literal = re.compile( r'\n::@xrst_line [0-9]+@\n\n' )
pattern_end_literal   = re.compile( r'\n[^ \t]')
#
# pattern_declare, pattern_use
# group(1): is the name of the label
# group(2): is the external destinatieon (empty for internal links).
# gruup(3): is the line where the label is defined or used.
pattern_declare = re.compile(
   r'\n\.\. _([^:\n]+):([^\n]*)@xrst_line ([0-9]+)@'
)
pattern_use = re.compile(
   r'`([^<\n]+)(<[^>\n]*>)`_[^\n]*@xrst_line ([0-9]+)@'
)
# ----------------------------------------------------------------------------
# {xrst_begin sphinx_label dev}
#
# Get Labels Declared Using Sphinx Commands
# #########################################
#
# {xrst_comment_ch #}
#
# Prototype
# *********
# {xrst_literal ,
#     # BEGIN_DEF, # END_DEF
#     # BEGIN_RETURN, # END_RETURN
# }
#
# data
# ****
# is the data for this page.
#
# page_file
# *********
# is the name of the xrst file containing the begin command for this page
# (only used for error reporting).
#
# page_name
# *********
# is the page name corresponding to *data*
# (only used for error reporting).
#
# m_external_label
# ****************
# For each label declared in *data* using sphinx commands,
# and that links to an external web site,
# *m_external_label* [ *label*.lower() ] is a match object
# for the label in *data* .
#
# m_internal_label
# ****************
# For each label declared in *data* using sphinx commands,
# and that links to a page in this web site,
# *internal* [ *label* ] is a match object
# for the label in *data* .
#
# Errors
# ******
# If two external labels have the same lower case value,
# an error is reported using :ref:`system_exit-name` .
#
# {xrst_end sphinx_label}
# BEGIN_DEF
def sphinx_label(data, page_file, page_name) :
   assert type(data) == str
   assert type(page_file) == str
   assert type(page_name) == str
   # END_DEF
   #
   # m_external_label, m_internal_label
   m_external_label  = dict()
   m_internal_label  = dict()
   #
   # literal_list
   literal_list = list()
   m_start = pattern_start_literal.search(data)
   while m_start :
      #
      # m_end
      m_end = pattern_end_literal.search(data, m_start.end() )
      #
      # literal_list
      if m_end == None :
         literal_list.append( (m_start.start(), len(data)) )
      else :
         literal_list.append( (m_start.start(), m_end.end()) )
      #
      # m_start
      m_start = pattern_start_literal.search( data, m_end.end() )
   #
   # pattern
   for pattern in [ pattern_declare, pattern_use ] :
      #
      # m_label
      m_label    = pattern.search(data)
      while m_label :
         #
         # skip
         skip = False
         for pair in literal_list :
            in_literal = pair[0] < m_label.start() and m_label.end() < pair[1]
            skip = skip or in_literal
         #
         if not skip :
            #
            # label_lower, line
            label       = m_label.group(1).strip(' ')
            label_lower = label.lower()
            destination = m_label.group(2).strip(' ')
            line        = m_label.group(3)
            #
            # msg
            msg = None
            for ending in [ '-name', '-title' ] :
               if label_lower.endswith(ending) :
                  msg  = f'Labels that end with {ending} are not allowed\n'
                  msg += 'because they are automatically generated by xrst\n'
                  msg += f'label.lower() = {label_lower}'
            if '@' in label_lower :
               msg  = f'Labels that use the @ character are not allowed\n'
               msg += 'because they are automatically generated by xrst\n'
               msg += f'label.lower() = {label_lower}'
            if label_lower == page_name :
               msg  = 'A label has the same lower case value as the page_name\n'
               msg += f'label.lower() = {label_lower}'
            if label_lower in m_external_label :
               previous_line = m_external_label[label_lower]
               msg  = 'A previous label has the same lower case value\n'
               msg += f'label.lower() = {label_lower}\n'
               msg += f'previous_line = {previous_line}'
            if msg != None :
               xrst.system_exit(msg,
                  file_name = page_file,
                  page_name = page_name,
                  m_obj     = m_label,
                  data      = data,
               )
            #
            # m_external_label, m_internal_label
            external =  destination != ''
            if external :
               m_external_label[label_lower] = m_label
            else :
               m_internal_label[label] = m_label
         #
         # m_label
         m_label = pattern.search(data, m_label.end())
   # BEGIN_RETURN
   #
   for result in [ m_external_label, m_internal_label ] :
      assert type(result) == dict
      for key in result.keys() :
         assert type( result[key] ) == re.Match
   return m_external_label, m_internal_label
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import tomli
import xrst
# {xrst_begin replace_spell dev}
# {xrst_comment_ch #}
#
# Replace spelling commands
# #########################
#
# tmp_dir
# *******
# is the directory where spell.toml is located
#
# spell.toml
# **********
# The file *tmp_dir* ``/spell.toml`` contains the information below.
# For each file that was included in the documentation,
# for each page in that file::
#
#     [file_name.page_name]
#     begin_line  = integer line number where begin command is located
#     start_spell = integer line number where the spell command starts
#     end_spell   = integer line number where the spell command ends
#     replace     = array of strings (words) that are not in dictionary
#
# It is called spell.tom because it is written by the
# :ref:`spell_cmd_dev-name` one page at a time.
#
# #.  file_name and page_name are strings.
# #.  file_name is relative to the
#     :ref:`config_file@directory@project_directory` .
# #.  Descriptions to the left (right) of the equal signs are literal text
#     (replaced by their values).
# #.  Line numbers start at one and are for the specified file.
# #.  The line number zero is used for start_spell and end_spell when
#     there is no spell command for this page.
# #.  The spell start and end lines do not overlap the begin line.
#
# Prototype
# *********
# {xrst_literal
#  # BEGIN_DEF
#  # END_DEF
#  }
#
# {xrst_end replace_spell}
#
# BEGIN_DEF
def replace_spell(tmp_dir) :
   assert type(tmp_dir) == str
   # END_DEF
   #
   # spell_toml
   file_obj  = open( f'{tmp_dir}/spell.toml' )
   file_data = file_obj.read()
   spell_toml = tomli.loads(file_data)
   #
   # file_name
   for file_name in spell_toml :
      #
      # page_list
      page_list = list()
      for page_name in spell_toml[file_name] :
         pair = (page_name , spell_toml[file_name][page_name] )
         page_list.append( pair )
      #
      # page_list
      order_fun    = lambda page_pair : page_pair[1]['begin_line']
      page_list = sorted(page_list, key = order_fun )
      #
      # data_in
      file_obj = open(file_name, 'r')
      data_in  = file_obj.read()
      file_obj.close()
      #
      # page_list
      # add begin_index, start_index, end_index
      line_number = 0
      for m_newline in re.finditer( r'(^|\n)',  data_in) :
         line_number += 1
         for page_name, page_info in page_list :
            if page_info['begin_line'] + 1 == line_number :
               page_info['begin_index'] = m_newline.start()
            if page_info['start_spell'] == line_number :
               page_info['start_index'] = m_newline.start() + 1
            if page_info['end_spell'] + 1 == line_number :
               page_info['end_index'] = m_newline.start() + 1
      #
      # data_copy
      data_copy = data_in
      data_copy = xrst.add_line_numbers(data_copy, file_name)
      #
      # data_out
      data_out      = ''
      data_in_index = 0
      #
      # page_name, page_info
      for page_name, page_info in page_list :
         #
         # m_begin
         m_begin = xrst.pattern['begin'].search(data_copy)
         while m_begin.group(3) != page_name :
            m_begin = xrst.pattern['begin'].search(data_copy, m_begin.end())
         #
         # m_end
         m_end = xrst.pattern['end'].search(data_copy, m_begin.end() )
         #
         # page_data
         page_data = data_copy[m_begin.end() : m_end.start() ]
         #
         # indent
         indent = xrst.auto_indent(page_data, file_name, page_name)
         #
         # comment_ch
         not_used, comment_ch = xrst.comment_ch_command(
            page_data, file_name, page_name
         )
         #
         # begin_line, indent_line
         if comment_ch :
            begin_line  = '\n' + indent + comment_ch + ' '
            indent_line = begin_line + 4 * ' '
         else :
            begin_line  = '\n' + indent
            indent_line = begin_line + 3 * ' '
         #
         # data_out, data_in_index
         begin_index   = page_info['begin_index']
         data_out     += data_in[data_in_index : begin_index]
         data_in_index = begin_index
         #
         # data_out
         replace = page_info['replace']
         if len(replace) > 0 :
            data_out += begin_line + '{xrst_spell'
            replace   = sorted(replace)
            i         = 0
            while i < len(replace) :
               word = replace[i]
               if i + 1 < len(replace) and word == replace[i+1] :
                  data_out += indent_line + word + ' ' + word
                  i         = i + 2
               else :
                  data_out += indent_line + word
                  i         = i + 1
            data_out += begin_line + '}'
         #
         # data_out, data_in_index
         if 'start_index' in page_info :
            start_index  = page_info['start_index']
            end_index    = page_info['end_index']
            #
            data_out     += data_in[data_in_index : start_index]
            data_in_index = end_index
      #
      # data_out
      data_out += data_in[data_in_index :]
      #
      # file_name
      file_obj = open(file_name, 'w')
      file_obj.write( data_out )
      file_obj.close()
   #
   return



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin suspend_cmd user}
{xrst_spell
   op
}

Suspend and Resume Commands
###########################

Syntax
******
- ``\{xrst_suspend}``
- ``\{xrst_suspend`` *boolean* ``}``
- ``\{xrst_suspend`` *left* *op* *right* ``}``
- ``\{xrst_resume}``

Discussion
**********
It is possible to suspend (resume) the xrst extraction and processing.
One begins (ends) the suspension with a line that only contains spaces,
tabs and a suspend command (resume command).
Each suspend command must have a corresponding resume command.

White Space
***********
Leading and trailing spaces in *boolean* , *left* , *op* , and *right*
are ignored.

boolean
*******
If the argument *boolean* is present, it must be true or false.
If is true, or not present, the section of input up to the next resume
is not included in the xrst extraction and processing for this page.
This argument is intended to be used by the
template command where it can be assigned the value true or false
for a particular expansion of the :ref:`template_cmd@template_file` .

*op*
****
This argument must be ``==`` or ``!=``.
It must be separated one or more spaces from *left* and *right* .

left, right
***********
If arguments *left* and *right* must not have any white space in them.
If

 *left* *op* *right*

is true (false) ,
the section of input up to the next resume
is not included in the xrst extraction and processing for this page.

Example
*******
:ref:`suspend_example-name`

{xrst_end suspend_cmd}
"""
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern_suspend, pattern_resume
pattern_suspend = re.compile(
   r'[^\\]{xrst_suspend *([^}]*)}'
)
pattern_resume  = re.compile(
   r'[^\\]{xrst_resume}'
)
# {xrst_begin suspend_cmd_dev dev}
# {xrst_comment_ch #}
#
# Remove text specified by suspend / resume pairs
# ###############################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page.
#
# page_file
# *********
# is the :ref:`begin_cmd@Page File` for this page.
#
# page_name
# *********
# is the name of this page.
#
# data_out
# ********
# The return data_out is a copy of data_in except that the text between
# and including each suspend / resume pair has been removed.
#
# {xrst_end suspend_cmd_dev}
# BEGIN_DEF
def suspend_command(data_in, page_file, page_name) :
   assert type(data_in) == str
   assert type(page_file) == str
   assert type(page_name) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_suspend
   m_suspend  = pattern_suspend.search(data_out)
   while m_suspend != None :
      #
      # m_resume
      m_resume      = pattern_resume.search(data_out, m_suspend.end())
      if m_resume == None :
         msg  = 'There is a suspend command without a '
         msg += 'corresponding resume command.'
         xrst.system_exit(msg,
            file_name=page_file,
            page_name=page_name,
            m_obj=m_suspend,
            data=data_out
         )
      #
      # m_obj
      m_obj = pattern_suspend.search(data_out, m_suspend.end())
      if m_obj != None :
         if m_obj.start() < m_resume.end() :
            msg  = 'There are two suspend commands without a '
            msg += 'resume command between them.'
            xrst.system_exit(msg,
               file_name=page_file,
               page_name=page_name,
               m_obj=m_obj,
               data=data_out
            )
      #
      # exclude
      argument = m_suspend.group(1).strip(' ')
      if argument in [ '', 'true' ] :
         exclude = True
      elif argument == 'false' :
         exclude = False
      else :
         argument = argument.split()
         if len(argument) != 3 :
            msg  = 'suspend command argument is not empty, true, false or'
            msg += 'three string separated by white space.'
            xrst.system_exit(msg,
               file_name=page_file,
               page_name=page_name,
               m_obj=m_obj,
               data=data_out
            )
         left  = argument[0]
         op    = argument[1]
         right = argument[2]
         if op == '==' :
            exclude = left == right
         elif op == '!=' :
            exclude = left != right
         else :
            msg  = '{' + 'xrst_suspend left op right}: '
            msg += 'op is not == or !=.'
            xrst.system_exit(msg,
               file_name=page_file,
               page_name=page_name,
               m_obj=m_obj,
               data=data_out
            )
      #
      # data_out
      if exclude :
         data_tmp  = data_out[: m_suspend.start() + 1]
         data_tmp += data_out[m_resume.end() : ]
      else :
         data_tmp  = data_out[: m_suspend.start() + 1]
         data_tmp += data_out[m_suspend.end() : m_resume.start() + 1]
         data_tmp += data_out[m_resume.end() : ]
      data_out  = data_tmp
      #
      # m_suspend
      m_suspend = pattern_suspend.search(data_out)
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
default_dict = dict()
r'''
{xrst_begin suspend_example}

Suspend Command Example
#######################

Discussion
**********
The project_name paragraph below
was taken from the xrst configure file documentation.

#. The documentation for the default table uses toml file format.
#. The python code that implements this default comes directly after
   and is not displayed in the documentation.

project_name
************
The only value in this table is the name of this project.
The default for this table is

{xrst_code toml}
[project_name]
data = 'project'
{xrst_code}

{xrst_suspend}'''
default_dict['project_name'] = { 'data' : 'project' }
r'''{xrst_resume}

xrst_code
*********
The file below uses ``xrst_code`` to display the toml version
of this default setting.

xrst_suspend
************
The file below uses ``xrst_suspend`` to avoid displaying the python version
of this default setting.

This Example File
*****************
{xrst_literal}


{xrst_end suspend_example}
'''



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
#
pattern_error = re.compile( r'@xrst_line *[0-9]+@[^\n]' )
# {xrst_begin remove_line_numbers dev}
# {xrst_spell
#     tuple
#     tuples
# }
# {xrst_comment_ch #}
#
# Remove the number numbers
# #########################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is a string with line number markers added by :ref:`add_line_numbers-name` .
# These lines number markers have the form:
#
#     ``@xrst_line`` *line_number* ``@`` .
#
# data_out
# ********
# The return data_out is a copy of data_in with the
# line number markers removed.
#
#
# rst2xrst_list
# *************
# The second return rst2xrst_list is a list of tuples.
# Each tuple in the list has two or four elements.
#
# First Tuple Element
# ===================
# is the line number in *data_out* corresponding to a line number marker
# that was removed from *data_in* .
# The lines in *data_out*  still contain the ``{xrst@before_title}`` commands
# that were in *data_in*.
# These are not included in the line number could (because they are
# removed before writing its rst file).
#
# Second Tuple Element
# ====================
# The second tuple element is the line number in the file that contains
# the begin command for this page.
#
# Third Tuple Element
# ===================
# This element is present If the current line in *data_out* is
# part of a template expansion.
# In this case, the third element is the template file name.
#
# Fourth Tuple Element
# ====================
# This element is present If the current line in *data_out* is
# part of a template expansion.
# In this case, the fourth element is the line in the template file.
#
# FourthTuple Element
# ===================
# If the current line in data_out corresponds to a template file,
# this is the line number in the template file. Otherwise, it is None.
#
# {xrst_end remove_line_numbers}
# BEGIN_DEF
def remove_line_numbers(data_in) :
   assert type(data_in) == str
   # END_DEF
   #
   # m_error
   m_error = pattern_error.search(data_in)
   if m_error :
      start = max(m_error.start() - 50, 0)
      end   = min(m_error.end() + 50, len(data_in))
      msg   = 'Program error: Line number tracking is confused:\n'
      msg  += 'Text before the bad line number =\n'
      msg  +=  data_in[start : m_error.start()]
      msg  += '\n---------------------------------\n'
      msg  += 'Line number with no newline at end =\n'
      msg  += m_error.group(0)
      msg  += '\n---------------------------------\n'
      msg  += 'Text after the bad line number =\n'
      msg  +=  data_in[m_error.end() :  end]
      msg  += '\n--------------------------------\n'
      xrst.system_exit(msg)
   #
   # template_list
   template_list = list()
   for m_template_begin in xrst.pattern['template_begin'].finditer(data_in) :
      template_start = m_template_begin.start()
      template_file  = m_template_begin.group(1).strip()
      page_line      = int( m_template_begin.group(2) )
      m_template_end = \
         xrst.pattern['template_end'].search(data_in, template_start)
      template_end   = m_template_end.end()
      #
      template_list.append(
         (template_start, template_end, page_line, template_file)
      )
   #
   # template_index
   def template_index(m_line) :
      assert type(m_line) == re.Match
      #
      if len(template_list) == 0 :
         return None
      #
      # index
      index = 0
      template_end  = template_list[index][1]
      while template_end < m_line.start() :
         index += 1
         if index == len(template_list) :
            return None
         template_end = template_list[index][1]
      template_start = template_list[index][0]
      if template_start <= m_line.start() and m_line.start() <= template_end :
         return index
      #
      return None
   #
   # previous_end
   # index of the end of the previous match
   previous_end  = 0
   #
   # line_out
   # index of next line in data_out
   line_out  = 1
   #
   # data_out, rst2xrst_list
   data_out      = ''
   rst2xrst_list = list()
   #
   # data_out, rst2xrst_list
   for m_line in xrst.pattern['line'].finditer(data_in) :
      #
      # before
      # character from end of previous match to start of this match
      before = data_in[previous_end  : m_line.start() ]
      #
      # line_out
      line_out  += before.count('\n')
      line_out  -= before.count('{xrst@before_title}\n')
      #
      # page_line, template_file, template_line
      index = template_index(m_line)
      if index == None :
         page_line     = int( m_line.group(1) )
         template_file = None
         template_line = None
      else :
         page_line     = template_list[index][2]
         template_line = int( m_line.group(1) )
         template_file = template_list[index][3]
      #
      # rst2xrst_list
      if index == None :
         rst2xrst_list.append( (line_out, page_line) )
      else :
         rst2xrst_list.append(
            (line_out, page_line, template_file, template_line)
         )
      #
      # data_out
      data_out += before
      #
      # previous_end
      previous_end = m_line.end()
   #
   # data_out
   data_out += data_in[previous_end  :]
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   assert type(rst2xrst_list) == list
   if 0 < len(rst2xrst_list) :
      assert type(rst2xrst_list[0]) == tuple
      assert type(rst2xrst_list[0][0]) == int
      assert type(rst2xrst_list[0][1]) == int
   return data_out, rst2xrst_list
   # END_RETURN
